package jess;

import java.util.*;
import java.io.*;

/**
 * Parent class of Defrules and Defqueries.
 * <P>
 * (C) 1997 Ernest J. Friedman-Hill and Sandia National Laboratories
 * @author Ernest J. Friedman-Hill
 */
public abstract class HasLHS extends Node
    implements Serializable, Visitable, Modular {
    String m_module;
    String m_name;
    String m_displayName;
    String m_docstring = "";
    private ArrayList m_nodes = new ArrayList();
    private Hashtable m_bindings = new Hashtable();
    ArrayList m_patts;
    int m_nodeIndexHash = 0;
    private StringBuffer m_compilationTrace;
    boolean m_new = true;
    boolean m_frozen = false;
    private HasLHS m_next = null;

    HasLHS(String name, String docstring, Rete engine) throws JessException {

        int colons = name.indexOf("::");
        if (colons != -1) {
            m_module = name.substring(0, colons);
            engine.verifyModule(m_module);
            m_name = name;
        } else {            
            m_module = engine.getCurrentModule();
            m_name = engine.resolveName(name);
        }

        int amp = m_name.indexOf('&');
        m_displayName = (amp == -1) ? m_name : m_name.substring(0, amp);

        m_docstring = docstring;
        m_patts = new ArrayList();
    }

    /**
     * Fetch the number of patterns on the LHS of this construct.
     * @return The number of patterns
     */
    public int getNPatterns() { return m_patts.size(); }
    /**
     * Return the idx-th Pattern on this construct's LHS.
     * @param idx The zero-based index of the desired pattern
     * @return the pattern
     */
    public Pattern getPattern(int idx) { return (Pattern) m_patts.get(idx); }

    /**
     * @param name 
     * @param factidx 
     * @param slotidx 
     * @param subidx 
     */
    /**
     * Return a string (useful for debugging) describing all the Rete network
     * nodes connected to this construct.
     * @return A textual description of all the nodes used by this construct
     */

    public String listNodes() {
        StringBuffer sb = new StringBuffer(100);
        for (int i=0; i< m_nodes.size(); i++) {
            sb.append(m_nodes.get(i));
            sb.append("\n");
        }
        return sb.toString();
    }

    HasLHS getNext() { return m_next; }
    void setNext(HasLHS next) { m_next = next; }
  
    void freeze(Rete engine) throws JessException {
        m_frozen = true;
    }

    void insertPatternAt(Pattern pattern, int index, Rete engine)
        throws JessException {
        ArrayList patterns = m_patts;
        m_bindings = new Hashtable();    
        m_patts = new ArrayList();
        patterns.add(index, pattern);
        for (int i=0; i<patterns.size(); i++)
            addPattern((Pattern) patterns.get(i), engine);
    }

    /**
     * Add a pattern to this construct
     */

    void addPattern(Pattern pattern, Rete engine) throws JessException {

        pattern = (Pattern) pattern.clone();
        Deftemplate dt = pattern.getDeftemplate();
        String varname = pattern.getBoundName();

        if (pattern.isBackwardChainingTrigger()) {
            if (varname == null)
                varname = RU.gensym("__factidx");
        }

        if (varname != null) {
            addBinding(varname, m_patts.size(), RU.PATTERN, -1, RU.FACT);
            pattern.setBoundName(varname);
        }


        m_patts.add(pattern);


        //Look for variables in the pattern, and create bindings for the
        //ones defined in this pattern.
        findVariableDefinitions(pattern, dt);

        // Now look for variables that were never given a proper definition --
        // those that were only negated.
        findUndefinedVariables(pattern);

        // Now handle '|' conjunctions by transforming tests into (or) Funcalls
        transformOrConjunctionsIntoOrFuncalls(pattern, dt, engine);

        // Add a 'not' so we don't fire if the fact exists during
        // backwards chaining
        if (pattern.isBackwardChainingTrigger()) {
            Pattern p =
                new Pattern(pattern, pattern.getNameWithoutBackchainingPrefix());
            p.setNegated(1);            
            addPattern(p, engine);
        }          
    }

    private void findVariableDefinitions(Pattern pattern,
                                         Deftemplate dt)
        throws JessException {

        for (int i=0; i< pattern.getNSlots(); i++) {
            for (int j=0; j< pattern.getNTests(i); j++) {
                Test1 test = pattern.getTest(i, j);
                Value val = test.m_slotValue;
                boolean eq = (test.m_test == Test1.EQ);
                if (val instanceof Variable) {
                    String name = val.variableValue(null);
                    if (m_bindings.get(name) == null) {
                        // Test for defglobal
                        if (name.startsWith("*"))
                            continue;
                                               
                        // If the test is positive, this is the
                        // definition of this variable, whether it's
                        // first or not, and whether it's defined in a
                        // "not" or not.

                        if (eq) {
                            int type = dt.getSlotDataType(i);
                            m_bindings.put(name,
                                           new BindingValue(name,
                                                            m_patts.size() - 1,
                                                            i,
                                                            test.m_subIdx,
                                                            type)); 
                        }
                    
                    }
                }
            }
        }
    }

    private void findUndefinedVariables(Pattern pattern) throws JessException {
        for (int i=0; i< pattern.getNSlots(); i++) {
            for (int j=0; j< pattern.getNTests(i); j++) {
                Test1 test = pattern.getTest(i, j);
                Value val = test.m_slotValue;
                boolean eq = (test.m_test == Test1.EQ);
                if (val instanceof Variable) {
                    String name = val.variableValue(null);
                    if (m_bindings.get(name) == null) {
                        // Ignore defglobals here
                        if (name.startsWith("*"))
                            ; 
                    
                        throw new JessException("HasLHS.addPattern",
                                                "First use of variable negated:",
                                                name);
                    }                
                }
            }
        }
    }

    private void transformOrConjunctionsIntoOrFuncalls(Pattern pattern,
                                                       Deftemplate dt,
                                                       Rete engine)
        throws JessException {

        for (int i=0; i< pattern.getNSlots(); i++) {
            int nTests = pattern.getNTests(i);

            if (nTests == 0)
                continue;
                                  
            // Rearrange the tests
            ArrayList tests = new ArrayList();
            
            int currentSubIndex = pattern.getTest(i, 0).m_subIdx;
            int doneIdx = 0;

            // This is a loop over sub-indexes in the test array. doneIdx
            // will be incremented inside the loop, and some constructs
            // will break out to this label.
        subIdxLoop:
            while (doneIdx < nTests) {
                // Find out if there are any ORs on this subslot
                boolean hasOrs = false;
                for (int j=doneIdx; j< nTests; j++) 
                    {
                        Test1 aTest = pattern.getTest(i, j);
                        if (aTest.m_subIdx != currentSubIndex)
                            break;
                        else if (aTest.m_conjunction == RU.OR) {
                            hasOrs = true;
                            break;
                        }
                    }
            
                // If no ORs on this subslot, just copy tests into ArrayList
                if (!hasOrs) {                
                    Test1 aTest = null;
                    for (int j=doneIdx; j< nTests; j++) {
                        aTest = pattern.getTest(i, j);
                        if (aTest.m_subIdx != currentSubIndex) {
                            currentSubIndex = aTest.m_subIdx;
                            continue subIdxLoop;
                        }
                        else {
                            tests.add(aTest);
                            ++doneIdx;                        
                        }
                    }                    
                    continue subIdxLoop;
                }
            
            
                // First find a variable to represent this (sub)slot; we
                // may have to create one
                Value var;
                Test1 firstTest = pattern.getTest(i, doneIdx);
            
                if (firstTest.m_slotValue.type() == RU.VARIABLE) {
                    var = firstTest.m_slotValue;
                    ++doneIdx;
                }
                else {
                    String name = RU.gensym(Tokenizer.BLANK_PREFIX);
                    var = new Variable(name, RU.VARIABLE);
                    m_bindings.put(name, 
                                   new BindingValue(name, m_patts.size() - 1,
                                                    i, currentSubIndex,
                                                    dt.getSlotDataType(i)));
                }
            
                tests.add(new Test1(Test.EQ, currentSubIndex, var));
            
                // We're going to build up this function call
                Funcall or = new Funcall("or", engine);
            
                // Count how many tests until an OR, so we can omit the
                // AND if not needed
                while (true) {
                    int andCount=1;
                    for (int j=doneIdx+1; j<nTests; j++) {
                        Test1 aTest = pattern.getTest(i, j);
                        if (aTest.m_conjunction == RU.OR ||
                            aTest.m_subIdx != currentSubIndex)
                            break;
                        else 
                            ++andCount;
                    }
                
                    if (andCount == 1) {
                        or.add(testToFuncall(pattern.getTest(i, doneIdx),
                                             var, engine));
                    }
                    else {
                        Funcall and = new Funcall("and", engine);
                        for (int j=doneIdx; j < doneIdx+andCount; j++)
                            and.add(testToFuncall(pattern.getTest(i, j),
                                                  var, engine));
                        or.add(new FuncallValue(and));
                    }
                
                    doneIdx += andCount;

                    if (doneIdx == nTests)
                        break;
                    else if (pattern.getTest(i, doneIdx).m_subIdx !=
                             currentSubIndex) 
                        break;
                }
                tests.add(new Test1(Test.EQ, currentSubIndex,
                                    new FuncallValue(or)));

                if (doneIdx < nTests &&
                    pattern.getTest(i, doneIdx).m_subIdx != currentSubIndex)
                    currentSubIndex = pattern.getTest(i, doneIdx).m_subIdx;
            }

            Test1[] testArray = new Test1[tests.size()];
            for (int j=0; j<testArray.length; j++)
                testArray[j] = (Test1) tests.get(j);
            pattern.replaceTests(i, testArray);
        }
    }

    /**
     *  Given a test, create an implied function call that does it
     */
    private Value testToFuncall(Test1 t, Value var, Rete engine)
        throws JessException {
        Value v = t.m_slotValue;
        switch (t.m_slotValue.type()) {
        case RU.FUNCALL: {
            if (t.m_test == Test.NEQ)
                return new FuncallValue(new Funcall("not", engine).arg(v));
            else
                return v;
        }
        default:        
            return new FuncallValue(new Funcall(t.m_test==Test.EQ ?
                                                "eq" : "neq", engine).
                arg(v).arg(var));
        }
    }

    private void addBinding(String name, int factidx, int slotidx,
                            int subidx, int type) 
        throws JessException {
        m_bindings.put(name,
                       new BindingValue(name, factidx, slotidx, subidx, type));
    }

    /**
     * @return 
     */
    Hashtable getBindings() { return m_bindings; }

    ArrayList getNodes() { return m_nodes; }
  
    void addNode(Node n) throws JessException {
        if (n == null)
            new JessException("HasLHS.addNode",
                              "Compiler fault",
                              "null Node added");

        for (int i=0; i<m_nodes.size(); ++i)
            if (n == m_nodes.get(i))
                return;

        appendCompilationTrace(n);

        ++n.m_usecount;
        m_nodes.add(n);

    }

    /**
     * Completely remove this construct from the Rete network, including
     * removing any internal nodes that are only used by this construct.
     * @param roots The roots of the Rete network where this construct lives.
     */
  
    void remove(Node root) throws JessException {
        Iterator e = m_nodes.iterator();
        while (e.hasNext()) {
            Node s = (Node) e.next();
            if (--s.m_usecount <= 0) {
                root.removeSuccessor(s);
                Iterator e2 = m_nodes.iterator();
                while (e2.hasNext()) {
                    Node n = (Node) e2.next();
                    n.removeSuccessor(s);
                }
            }
        }
        m_nodes.clear();
    }  

    private void appendCompilationTrace(Node n) {
        if (m_compilationTrace == null)
            m_compilationTrace = new StringBuffer(m_name + ": ");

        if (n.m_usecount == 0)
            m_compilationTrace.append("+");
        else
            m_compilationTrace.append("=");

        m_compilationTrace.append(n.getCompilationTraceToken());
    }

    String getCompilationTraceToken() { return "t"; }

    StringBuffer getCompilationTrace() { return m_compilationTrace; }

    /**
     * Set the node-index-hash of this construct. The node-index-hash
     * value effects the indexing efficiency of the join nodes for this
     * construct. Larger values will make constructs with many partial
     * matches faster (to a point). Must be set before construct is
     * added to engine (so is typically set during parsing via the
     * equivalent Jess command. 
     @param h The node index hash value
    */

    public void setNodeIndexHash(int h) { m_nodeIndexHash = h; }

    /**
     * Get the node-index-hash setting of this construct.
     * @return The node-index-hash of this construct
     */
    public int getNodeIndexHash() { return m_nodeIndexHash; }

    /**
     * Fetch the name of this construct
     * @return The name of this construct
     */
    public final String getName() { return m_name; }

    /**
     * Fetch the display name of this construct
     * @return The display name of this construct
     */
    public String getDisplayName() {
        return m_displayName;
    }
  
    /**
     * Get the documentation string for this construct.
     * @return The docstring for this construct
     */
    public final String getDocstring() { return m_docstring; }

    // Compiler calls this after we've had initial update
    void setOld() { m_new = false;}

    public abstract Object accept(Visitor jv);

    public String getModule() {
        return m_module;
    }
}



