package jess;

import java.util.*;
import java.io.*;

/** **********************************************************************
 * Generates a pattern network
 *
 * See the paper
 * "Rete: A Fast Algorithm for the Many Pattern/ Many Object Pattern
 * Match Problem", Charles L.Forgy, Artificial Intelligence 19 (1982), 17-37.
 *
 * The implementation used here does not follow this paper; the present
 * implementation models the Rete net more literally as a set of networked Node
 * objects with interconnections.
 *
 * $Id: ReteCompiler.java,v 1.11 2001/12/07 02:26:20 ejfried Exp $
 * (C) 1998 E.J. Friedman-Hill and the Sandia Corporation
 ********************************************************************** */


class ReteCompiler implements Serializable
{

  private int m_hashkey = 101;
  public int getHashKey() { return m_hashkey; }
  public void setHashKey(int h) { m_hashkey = h; }
  
  /**
    The roots of the pattern network
    */

  private RootNode m_root = new RootNode();

  /**
   * @return 
   */
  public final Node getRoot() { return m_root; }

  public ReteCompiler() {
  }

  /**
   * Evaluate a funcall, replacing all variable references with
   * BindingValues.  Bind each variable to the first occurrence in
   * rule.
   * @param table 
   * @param v 
   * @param dr 
   * @exception JessException 
   * @return */
  private Value eval(Hashtable table, Value v, HasLHS dr)
       throws JessException 
  {
    if (v.type() != RU.FUNCALL)
      return v;
    
    Funcall vv = (Funcall) v.funcallValue(null).clone();
    for (int i=0; i<vv.size(); i++) 
      {
        Value val = vv.get(i);
        if (val instanceof Variable)
          {
            String name = val.variableValue(null);
            BindingValue b = (BindingValue) table.get(name);
            if (b != null)
              {
                vv.set(b, i);
              }
            else
              {
                // Defglobal
                if (name.charAt(0) != '*') {
                  System.out.println(new PrettyPrinter(dr));
                  compilerError("eval", "Unbound variable found in funcall: " +
                                name + " parsing defrule " + dr.getName());
                }
              }
          }
        else if (val.type() ==  RU.FUNCALL) 
          {
            // nested funcalls
            vv.set(eval(table, val, dr), i);
          }
      }

    return new FuncallValue(vv);
  }

  /**
   * Call this on a funcall value AFTER eval has modified it.
   * it returns true iff this funcall contains binding to patterns other
   * than the one named by currentFactNumber
   * @param v 
   * @param currentFactNumber
   * @exception JessException 
   * @return 
   */
  private boolean checkForMultiPattern(Value v, int currentFactNumber)
       throws JessException 
  {
    ValueVector vv = v.funcallValue(null);

    for (int i=1; i<vv.size(); i++) 
      {
        Value arg = vv.get(i);
        if (arg.type() ==  RU.BINDING &&
            ((BindingValue) arg).getFactNumber() != currentFactNumber)
          return true;
        
        else if (arg.type() ==  RU.FUNCALL &&
                 checkForMultiPattern(arg, currentFactNumber)) 
            return true;
      }
    return false;

  }


  private Hashtable m_doneVars = new Hashtable();
  /**
   * Add a rule's patterns to the network
   * @param r 
   * @exception JessException 
   */  
  public synchronized void addRule(HasLHS r, Rete engine)
      throws JessException 
  {
    int hash = r.getNodeIndexHash();
    if (hash == 0)
      hash = m_hashkey;

    // Tell the rule to fix up its binding table; it must be complete now.
    r.freeze(engine);

    // Fetch the rule's binding table; no more makeVarTable!
    Hashtable table = r.getBindings();

    // 'terminals' will be where we hold onto the final links in the
    // chain of nodes built during the first pass for each pattern
    Node[] terminals = new Node[r.getNPatterns()];
    Node[] ruleRoots = new Node[r.getNPatterns()];
    
    /* *********
       FIRST PASS
       ********* */

    // In the first pass, we just create some of the one-input
    // nodes for each pattern

    // These one-input nodes compare a certain slot in a fact (token) with a
    // fixed, typed, piece of data. Each pattern gets one special
    // one-input node, the TECT node, which checks the class type
    // and class name.


    // iterate over each of the rule's patterns
    for (int i=0; i<r.getNPatterns(); i++) 
      {

        // get this pattern
        Pattern p = (Pattern) r.getPattern(i);

        // And its deftemplate
        Deftemplate deft = p.getDeftemplate();

        // If this is a 'test' CE, we have to treat it slightly differently
        boolean isTest = (p.getName().equals("test"));

        ////////////////////////////////////////////////////////////
        // Every pattern must have a definite class name
        // Therefore, the first node in a chain always
        // checks the class name of a token (except for test CEs)
        ////////////////////////////////////////////////////////////

        Node last = null;
        
        last = m_root.mergeSuccessor(
               isTest? (Node1) new Node1NONE() : (Node1) new Node1TECT(p.getName()),r);
        
        ruleRoots[i] = last;

        // First we have to find all the multifields, because these change the
        // 'shape' of the facts as they go through the net. 

        for (int slot=0; !isTest && slot<p.getNSlots(); slot++) 
          {
            
            // any tests on this slot?
            if (deft.getSlotType(slot) != RU.MULTISLOT)
              continue;

            // No multifields; just test length
            if (p.getNMultifieldsInSlot(slot) == 0)
              {
                int length = p.getSlotLength(slot);
                if (length != -1)
                  last = last.mergeSuccessor( new Node1MTELN(slot, length), r);
              }

            // Multifields that needs separating out?
            else 
              last = last.mergeSuccessor( new Node1MTMF(slot,
                                                        p.getMultifieldFlags(slot)), r);
          }
        
        
        ////////////////////////////////////////////////////////////
        // Simplest basic tests are done here
        ////////////////////////////////////////////////////////////

        for (int j=0; !isTest && j<p.getNSlots(); j++) 
          {

            // any tests on this slot?
            if (p.getNTests(j) == 0)
              continue;

            int testIdx = j;
        
            for (int k=0; k<p.getNTests(j); k++) 
              {
                Test1 test = p.getTest(j, k);
                Value slotValue = test.m_slotValue;


                if (slotValue instanceof Variable)
                  {
                    // Single tests against only global variables
                    String name = slotValue.variableValue(null);
                    if (!(name.startsWith("*") && name.endsWith("*")))
                      continue;
                  }
                
                // expand variable references in funcalls to index,
                // slot, subslot triples
                Value v =  eval(table, slotValue, r);

                if (slotValue.type() != RU.FUNCALL || !checkForMultiPattern(v, i))
                  last = addSimpleTest(last, r, testIdx, test, v, engine);
              }
          }
        terminals[i] = last;
      }

    /* *********
       SECOND PASS
       ********* */

    // In this pass, we are looking for variables which must be
    // instantiated the same way twice in one fact. IE, the pattern look like
    // (foo ?X foo ?X), and we're looking for facts like (foo bar foo bar).
    // NOT versions are handled as well.

    // iterate over each of the rule's patterns
    for (int i=0; i<r.getNPatterns(); i++) 
      {
          
        // get this pattern
        Pattern p = (Pattern) r.getPattern(i);

        // If this is a 'test' CE, we have to treat it slightly differently
        if (p.getName().equals("test"))
          continue;

        // workspace to track variables that have been done.
        m_doneVars.clear();

        // find a variable slot, if there is one. If one is found,
        // look at the rest of
        // the fact for another one with the same name.
        // If one is found, create the
        // appropriate node and put it in place.

        // NOTs make things a bit more complex.
        // There are a few cases for a varname
        // appearing twice in a pattern:

        // ?X ?X        ->        generate a TEV1 node.
        // ?X ~?X       ->        generate a TNEV1 node.
        // ~?X ?X       ->        generate a TNEV1 node.
        // ~?X ~?X      ->        (DO NOTHING!)


        // look for a slot in the pattern containing a variable

        for (int j= 0; j < p.getNSlots(); j++) 
          {
            // any tests for this slot?
            if (p.getNTests(j) == 0)
              continue;

            for (int k= 0; k < p.getNTests(j); k++) 
              {
                Test1 test_jk = p.getTest(j, k);                
                if (!(test_jk.m_slotValue instanceof Variable))
                  continue;
                
                // see if we've done this one before.
                String varName = test_jk.m_slotValue.variableValue(null);
                if (m_doneVars.get(varName) != null)
                  continue;
                
                // no, we haven't. Find each other occurrence.
                // We start the search at the same slot since it might be a
                // multislot!
                for (int n=j; n < p.getNSlots(); n++) 
                  {
                    if (p.getNTests(n) == 0)
                      continue;
                    for (int o= 0; o < p.getNTests(n); o++) 
                      {
                        Test1 test_no = p.getTest(n, o);                

                        // This can happen since we're researching the same slot.
                        if (test_no == test_jk)
                          continue;
                        if (test_no.m_slotValue instanceof Variable &&
                            test_no.m_slotValue.equals(test_jk.m_slotValue))
                          {
                            // we've identified another slot with the same variable.
                            addMultipleReferenceTest(terminals, i, 
                                                     j, test_jk,
                                                     n, test_no,
                                                     r);
                          }
                      }
                  }
                m_doneVars.put(varName, varName);
              }
          }
        
      } // end of second pass
    
    // Attach a NodeRTL to the very first tail, so this one will do a LEFT calltype
    // instead of a RIGHT one.

    terminals[0] = terminals[0].mergeSuccessor(new Node1RTL(), r);
    if (r.getPattern(0).getLogical()) 
        ((Defrule) r).setLogicalInformation((Node1RTL) terminals[0], 1);

    /* *********
       THIRD PASS
       ********* */

    // Now we start making some two-input nodes. These nodes check that
    // a variable with the same name in two patterns is instantiated the
    // same way in each of two facts; or not, in the case of negated
    // variables. An important thing to remember: the first instance of a
    // variable can never be negated. We'll check that here and throw
    // an exception if it's violated.  We can compare every other instance
    // of the variable in this rule against this first one - this simplifies
    // things a lot! We'll use simplified logic which will lead to a few
    // redundant, but correct tests.

    // Two-input nodes can contain many tests, so they are rather more
    // complex than the one-input nodes. To share them, what we'll do is build
    // a new node, then compare this new one to all possible shared ones.
    // If we can share, we just throw the new one out. The inefficiency is
    // gained back in spades at runtime, both in memory and speed. Note that
    // NodeNot2 type nodes cannot be shared.

    /*
      The number of two-input nodes that we create is *determinate*: it is
      always one less than the number of patterns. For example, w/ 4 patterns,
      numbered 0,1,2,3, and the following varTable:
      (Assuming RU.SLOT_SIZE = 2)

      <PRE>

      X  Y  N
      0  1  2
      2  4  4

      </PRE>
      generated from the following rule LHS:

      <PRE>
      (foo ?X ?X)
      (bar ?X ?Y)
      (Goal (Type Simplify) (Object ?N))
      (Expression (Name ?N) (Arg1 0) (Op +) (Arg2 ~?X))
      </PRE>

      Would result in the following nodes being generated
      (Assuming SLOT_DATA == 0, SLOT_TYPE == 1, SLOT_SIZE == 2):

      <PRE>
        0     1
         \   /
      ___L___R____
      |          |            2
      | 0,2 = 2? |           /
      |          |          /
      ------------ \0,1    /
                   _L______R__                3
                  |          |               /
                  | NO TEST  |              /
                  |          |             /
                  ------------ \0,1,2     /
                                L_______R__
                                | 0,2 != 8?|
                                | 2,4 = 2? |
                                |          |
                                ------------
                                     |0,1,2,3
                                     |
                                (ACTIVATE)

      <PRE>

      Where the notation 2,4 = 8? means that this node tests tbat index 4 of
      fact 2 in the left token is equal to index 8 in the right
      token's single fact. L and R indicate Left and Right inputs.
      */

    // for each pattern, starting with the second one

    for (int i=1; i < (r.getNPatterns()); i++) 
      {

        // get this pattern
        Pattern p = (Pattern) r.getPattern(i);

        // Keep track of which variables we've tested already for this pattern.
        m_doneVars.clear();

        // If this is a 'test' CE, we have to treat it slightly differently
        boolean isTest = (p.getName().equals("test"));

        // construct an appropriate 2 input node...
        NodeTest n2;
        
        if (isTest)
          n2 = new NodeTest(engine);
        else if (p.getNegated() != 0)
          n2 = new NodeNot2(engine, hash);
        else
          n2 = new Node2(engine, p.getUnique(), hash);
        
        // now tell the node what tests to perform

        // for each field in this pattern
        for (int j=0; j< p.getNSlots(); j++) 
          {

            // any tests for this slot?
            if (p.getNTests(j) == 0)
              continue;

            // for every test on this slot..
            for (int k=0; k< p.getNTests(j); k++) 
              {
                Test1 test_jk = p.getTest(j, k);
                // if this test is against a variable...
                if (test_jk.m_slotValue instanceof Variable) 
                  {
                    // find this variable in the table
                    String name = p.getTest(j, k).m_slotValue.variableValue(null);

                    if (m_doneVars.get(name) != null)
                      continue;

                    BindingValue b = (BindingValue) table.get(name);
                    if (b == null)
                      compilerError("addRule",
                                    "Corrupted VarTable: var " +
                                    name + " not in table");


                    // if this is the first appearance, no test.
                    else if (b.getFactNumber() == i)
                      continue;

                    if (test_jk.m_test == Test.EQ)
                      n2.addTest(Test.EQ,
                                 b.getFactNumber(),
                                 b.getSlotIndex(),
                                 b.getSubIndex(),
                                 j,
                                 test_jk.m_subIdx);
                    else
                      n2.addTest(Test.NEQ,
                                 b.getFactNumber(),
                                 b.getSlotIndex(),
                                 b.getSubIndex(),
                                 j,
                                 test_jk.m_subIdx);

                    m_doneVars.put(name, name);

                    // if this test is a function call
                  }
                else if (test_jk.m_slotValue.type() == RU.FUNCALL) 
                  {
                    // This simplifies things somewhat!
                    if (p.getDeftemplate().getBackwardChaining())
                      throw new JessException("ReteCompiler.addRule",
                                              "Can't use funcalls in backchained patterns",
                                              p.getName());
                    
                    // expand the variable references to index, slot pairs
                    // we do this again even though we did it in pass one
                    // we don't want to destroy the patterns themselves
                    // Tell Eval to bind variables to first occurrence in Rule.
                    Value v = eval(table, test_jk.m_slotValue, r);

                    // if other facts besides this one are mentioned, generate a test

                    if (isTest || checkForMultiPattern(v, i)) {
                        
                        if (isTest && p.getNegated() % 2 == 1) {
                            Funcall not = new Funcall("not", engine);
                            not.arg(v);
                            v = new FuncallValue(not);
                        }

                        if (test_jk.m_test == Test.EQ)
                          n2.addTest(Test.EQ, j,
                                     test_jk.m_subIdx, v);
                        else
                          n2.addTest(Test.NEQ, j,
                                     test_jk.m_subIdx, v);
                      }
                  }
              }
          }
      
        // search through the successors of this pattern and the next one.
        // Do they have any in common, and if so, are they equivalent to the
        // one we just built? If so, we don't need to add the new one!
        // Don't share NodeNot2's.
        
        n2 = addJoinNode(n2, terminals, i, r);
                    
        if (p.getDeftemplate().getBackwardChaining() && ! p.getExplicit())
          {
            // n2 can't be a NodeTest, since do-backwards-chaining would
            // have rejected the call. Explicit CEs are those for which we should
            // never do backwards chaining
            ((Node2) n2).setBackchainInfo(p, r);
          }

        // Safe because logical is disallowed in defqueries, and
        // (test) is disallowed in (logical.)
        if (p.getLogical())
            ((Defrule) r).setLogicalInformation((Node2) n2, i+1);
            
        // Handling nested nots. We have already built the innermost "not".
        // Now we add the outer ones. Each not feeds into the -right- input of
        // the next outermost one. Only the innermost has any tests in it.
        if (p.getNegated() > 1)
          {
            for (int nested_nots=1; nested_nots < p.getNegated(); nested_nots++)
              {
                Node nltr = new Node1LTR();  // A left-to-right input adapter
                NodeTest not = new NodeNot2(engine, hash);   // outer not
                not.complete();              // must call this after all tests added
                n2.addSuccessor(nltr, r);  // the inner not feeds into the adapter
                nltr.addSuccessor(not, r); // the adapter feeds into the outer not
                // the inner not's left input also goes into outer
                terminals[i-1].addSuccessor(not, r);  
                n2 = not;
              }
            
          }

        // Advance the tails
        terminals[i-1] = terminals[i] = n2;

      }
    
    /* ************
       FOURTH PASS
       ************ */

    // All that's left to do is to create the terminal node.
    // This is very easy.

    terminals[r.getNPatterns() - 1].mergeSuccessor(r, r);
    //Tell the engine to update this rule if the fact list isn't empty
    Hashtable uniqueRoots = new Hashtable();
    for (int i=0; i<ruleRoots.length; i++)
      uniqueRoots.put(ruleRoots[i], ruleRoots[i]);
    try {
        engine.updateNodes(uniqueRoots);    
    } finally {
        r.setOld();
    }
  }

  private NodeTest addJoinNode(NodeTest n2, Node[]terminals, int i, HasLHS r)
    throws JessException
  {

    // Find any node equal to this one, or return this one.
    NodeTest new1 = (NodeTest) terminals[i-1].resolve(n2);

    // If we found a preexisting one...
    if (new1 != n2)
      {
        NodeTest new2 = (NodeTest) terminals[i].resolve(n2);

        // If the two nodes are the same node, these two tails are
        // already connected.
        if (new1 == new2)
          {
            r.addNode(new1);
            return new1;
          }        
      }
    
    // Here we use add, not merge. We only want to share if the two nodes were
    // already connected.
    n2.complete();
    terminals[i-1].addSuccessor(n2, r);
    terminals[i].addSuccessor(n2, r);
    return n2;
  }
  

  private Node addSimpleTest(Node last, HasLHS r,
                             int testIdx, Test1 test, Value v, Rete engine)
    throws JessException
  {
    Node1 node;
    switch(test.m_test)
      {
      case Test1.EQ:
        switch (test.m_subIdx)
          {
          case -1:
            node = new Node1TEQ(testIdx, v, engine);
            break;
          default:
            node = new Node1MTEQ(testIdx, test.m_subIdx, v, engine);
            break;
          }
        break;
      default:
        switch (test.m_subIdx)
          {
          case -1:
            node = new Node1TNEQ(testIdx, v, engine);
            break;
          default:
            node = new Node1MTNEQ(testIdx, test.m_subIdx, v, engine);
            break;
          }
        break;
      }
    
    return last.mergeSuccessor(node, r);                    
  }
    
  // Add a test for agreement among multiple references to the same
  // variable inside one pattern.

  private void addMultipleReferenceTest(Node[] terminals, int i, 
                                        int slot1, Test1 test_jk,
                                        int slot2, Test1 test_no,
                                        HasLHS r)
    throws JessException
  {
    if (test_jk.m_test == Test.EQ) 
      {
        if (test_no.m_test == Test.EQ)
          terminals[i] =
            terminals[i].mergeSuccessor(
                                        new Node1TEV1(slot1, test_jk.m_subIdx,
                                                      slot2, test_no.m_subIdx),
                                        r);
        else
          terminals[i] =
            terminals[i].mergeSuccessor(
                                        new Node1TNEV1(slot1, test_jk.m_subIdx,
                                                       slot2, test_no.m_subIdx),
                                        r);
        
      }
    else 
      {
        if (test_no.m_test == Test.EQ)
          terminals[i] =
            terminals[i].mergeSuccessor(
                                        new Node1TNEV1(slot1, test_jk.m_subIdx,
                                                       slot2, test_no.m_subIdx),
                                        r);
        else
          ;
      }

  }


  private void compilerError(String routine, String message)
      throws RuleCompilerException {
      throw new RuleCompilerException("ReteCompiler." + routine, message);
  }
    
}

class RootNode extends Node1
{
  boolean callNodeRight(Token t) throws JessException
  {
    passAlong(t);
    return true;
  }
}
